import { assignDeepOmitBlank, ChartMeasureRoleType, ChartOrient, cloneDeep, getChartCategory, getChartCategory2, getChartSeries, getEntityHierarchy, getEntityProperty, getPropertyCaption, getPropertyHierarchy, getPropertyMeasure, mergeOptions, omitBlank } from '@metad/ocap-core';
import { LinearGradient } from 'echarts/lib/util/graphic';
import { assign, includes, indexOf, isArray, isEmpty, isNil, maxBy, minBy, range } from 'lodash-es';
import { axisOrient, getCategoryAxis, getMeasureAxis, getValueAxis } from './axis';
import { getChromaticScale } from './chromatics';
import { formatMeasureNumber } from './common';
import { coordinates, gatherCoordinates } from './coordinates';
import { DecalPatterns } from './decal';
import { referenceLines } from './series';
import { stackedForMeasure } from './stacked';
import { AxisEnum, isCoordinatePolar } from './types';
import { getColorLightnessVisualMap } from './visualMap';
export function cartesian(data, chartAnnotation, entityType, settings, options, type) {
    const [categoryAxis, valueAxis] = axisOrient(chartAnnotation.chartType.orient);
    const context = {
        data,
        entityType,
        settings,
        options,
        chartAnnotation,
        categoryAxis,
        valueAxis
    };
    const coordinateDatas = coordinates(context, type === 'pie' ? 'pie' : 'grid', cartesianCoordinate);
    return {
        ...context,
        echartsOptions: gatherCoordinates(coordinateDatas, type, options)
    };
}
/**
 * 单个笛卡尔坐标系, 对应一个 grid 内的组件
 *
 * @param data 数据数组
 * @param chartAnnotation 图形注解
 * @param entityType 实体类型
 * @param settings 图形设置
 * @param options 图形属性
 * @returns
 */
export function cartesianCoordinate(context, data) {
    const { chartAnnotation, entityType, settings, options } = context;
    const category = getChartCategory(chartAnnotation);
    const categoryProperty = getEntityProperty(entityType, category);
    const category2 = getChartCategory2(chartAnnotation);
    const chartSeries = getChartSeries(chartAnnotation);
    let datasets = [];
    if (chartSeries) {
        datasets = chartAnnotation.measures
            .filter(({ role }) => !includes([ChartMeasureRoleType.Tooltip, ChartMeasureRoleType.Size, ChartMeasureRoleType.Lightness], role))
            .map((measure) => {
            return stackedForMeasure(data, measure, chartAnnotation, entityType, settings, options);
        });
    }
    else {
        const seriesComponents = measuresToSeriesComponents(chartAnnotation.measures, data, entityType, settings);
        data = [...data];
        // Fill up measure null value
        if (data[0]) {
            data[0] = { ...data[0] };
            chartAnnotation.measures.forEach((measure) => {
                if (isNil(data[0][measure.measure])) {
                    data[0][measure.measure] = '-';
                }
            });
        }
        datasets = [
            {
                dataset: {
                    source: data
                },
                seriesComponents
                // tooltip: getEChartsTooltip(
                //   options?.tooltip,
                //   categoryProperty,
                //   chartAnnotation.measures.map((measure) => ({
                //     measure,
                //     property: getEntityProperty(entityType, measure)
                //   })),
                //   seriesComponents,
                //   settings.locale
                // )
            }
        ];
    }
    context.datasets = datasets;
    const type = chartAnnotation.chartType.variant === 'polar' ? 'Polar' : 'Cartesian2d';
    const { categoryAxis, valueAxis } = type === 'Polar'
        ? getPolarCoordinateSystem(context, data, settings.locale)
        : getCoordinateSystem(context, data, settings.locale);
    const gridOptions = {
        type,
        name: '',
        grid: mergeOptions({
            top: 0,
            right: 0,
            bottom: 0,
            left: 0,
            containLabel: true
        }, options?.grid),
        [categoryAxis.orient]: categoryAxis.axis,
        [valueAxis.orient]: valueAxis.axis,
        visualMap: [],
        datasets: [],
        tooltip: [],
        legend: []
    };
    if (isCoordinatePolar(gridOptions)) {
        gridOptions.polar = mergeOptions({}, options?.polar);
    }
    datasets.forEach(({ dataset, seriesComponents }) => {
        gridOptions.datasets.push({
            dataset,
            series: seriesComponents.map((seriesComponent) => {
                const { series, visualMaps } = serializeSeriesComponent(dataset, seriesComponent, entityType, valueAxis, settings, options);
                gridOptions.visualMap.push(...visualMaps);
                if (!isNil(seriesComponent.valueAxisIndex)) {
                    series[valueAxis.orient + 'Index'] = seriesComponent.valueAxisIndex;
                }
                /**
                 * 堆积系列情况下不适用 encode, 使用的应该是 seriesLayoutBy: "row" name: "[Canada]" 与 dataset 中的行数据进行的对应
                 */
                if (isNil(series.seriesLayoutBy)) {
                    series.encode = {
                        [AxisEnum[categoryAxis.orient]]: getPropertyHierarchy(category),
                        [AxisEnum[valueAxis.orient]]: category2 ? getPropertyHierarchy(category2) : seriesComponent.measure,
                        tooltip: seriesComponent.tooltip
                    };
                    if (!settings?.universalTransition) {
                        series.encode.itemName = getPropertyCaption(categoryProperty);
                    }
                }
                // series.tooltip = tooltip
                /**
                 * 解决 valueFormatter 与 formatter 里 member caption 共存的问题:
                 * encode.itemName 决定 name, valueFormatter 决定 value.
                 */
                series.tooltip = {
                    valueFormatter: (value) => {
                        return formatMeasureNumber({ measure: seriesComponent, property: seriesComponent.property }, value, settings.locale, seriesComponent.formatting?.shortNumber);
                    }
                };
                return assignDeepOmitBlank(assignDeepOmitBlank(series, options?.seriesStyle, 2), seriesComponent.chartOptions?.seriesStyle, 2);
            })
        });
    });
    // visualMaps
    if (options?.visualMaps?.length) {
        gridOptions.visualMap = gridOptions.visualMap ?? [];
        options.visualMaps.forEach((item, index) => {
            gridOptions.visualMap[index] = mergeOptions(gridOptions.visualMap[index], cloneDeep(item));
        });
    }
    if (options?.tooltip?.trigger === 'axis') {
        const measure = getMeasureAxis(chartAnnotation, ChartMeasureRoleType.Axis1, 0);
        gridOptions.tooltip.push({
            ...omitBlank(options.tooltip),
            trigger: 'axis',
            valueFormatter: (value) => {
                return formatMeasureNumber({ measure, property: null }, value, settings.locale, measure.formatting?.shortNumber);
            }
        });
    }
    else {
        gridOptions.tooltip.push({
            ...omitBlank(options?.tooltip),
            trigger: 'item'
        });
    }
    // legend
    if (options?.legend) {
        gridOptions.legend.push(mergeOptions({
            formatter: (name) => {
                let label = name;
                gridOptions.datasets.find(({ series }) => series.find((item) => {
                    if (item.name === name) {
                        label = item.caption;
                        return true;
                    }
                    return false;
                }));
                return label || name;
            }
        }, options.legend));
    }
    return gridOptions;
}
/**
 * Serialize measure to echarts series component options
 *
 * @param dataset
 * @param seriesComponent
 * @param entityType
 * @param valueAxis
 * @param settings
 * @param options
 * @returns
 */
export function serializeSeriesComponent(dataset, seriesComponent, entityType, valueAxis, settings, options) {
    const visualMaps = [];
    const series = {
        member: seriesComponent.member,
        id: seriesComponent.id,
        name: seriesComponent.name ?? seriesComponent.id,
        caption: seriesComponent.caption,
        type: seriesComponent.seriesType,
        measure: seriesComponent.measure,
        datasetIndex: seriesComponent.datasetIndex,
        seriesLayoutBy: seriesComponent.seriesLayoutBy,
        stack: seriesComponent.seriesStack,
        universalTransition: settings?.universalTransition
    };
    if (seriesComponent.palette?.name || seriesComponent.palette?.color) {
        const visualMap = {
            show: false,
            type: 'continuous',
            inRange: {},
            text: [seriesComponent.property?.label]
        };
        const domain = seriesComponent.domain || [seriesComponent.dataMin, seriesComponent.dataMax];
        if (seriesComponent.palette.name) {
            const colorFun = getChromaticScale(seriesComponent.palette.name, [0, seriesComponent.dataSize - 1], seriesComponent.palette.reverse);
            visualMap.inRange.color = range(seriesComponent.dataSize).map((i) => colorFun(i));
        }
        else if (seriesComponent.palette.color) {
            visualMap.inRange.color = seriesComponent.palette.color;
        }
        visualMap.min = domain[0];
        visualMap.max = domain[1];
        visualMap.dimension = getVisualMapValueAxisIndex(dataset, seriesComponent, null);
        series.color = getLegendColorForVisualMap(visualMap.inRange.color);
        visualMaps.push(visualMap);
    }
    if (!isNil(seriesComponent.palette?.pattern)) {
        series.itemStyle = series.itemStyle ?? {};
        series.itemStyle.decal = DecalPatterns()[seriesComponent.palette.pattern];
    }
    if (!isEmpty(seriesComponent.referenceLines) && valueAxis) {
        assign(series, referenceLines(seriesComponent, null, options, valueAxis.orient, settings?.locale));
    }
    if (seriesComponent.sizeMeasure) {
        visualMaps.push(getSymbolSizeVisualMap(seriesComponent.sizeMeasure, getEntityProperty(entityType, seriesComponent.sizeMeasure), dataset, options?.visualMaps));
    }
    if (seriesComponent.lightnessMeasure) {
        const lightnessProperty = getEntityProperty(entityType, seriesComponent.lightnessMeasure);
        visualMaps.push(getColorLightnessVisualMap(seriesComponent.lightnessMeasure, lightnessProperty, dataset, options?.visualMaps));
    }
    if (options?.seriesStyle?.endLabel?.show) {
        series.endLabel = {
            show: true,
            formatter: (params) => {
                const member = params.seriesId.slice(0, -seriesComponent.measure.length);
                const label = dataset.series.find((item) => item.value === member)?.label;
                const index = params.dimensionNames.findIndex((item) => item === member);
                const value = params.value[index + 1];
                return (label ?? member) + ': ' + formatMeasureNumber({ measure: seriesComponent, property: seriesComponent.property }, value, settings?.locale);
            }
        };
    }
    return {
        series,
        visualMaps
    };
}
export function getVisualMapValueAxisIndex(dataset, seriesComponent, series) {
    if (series && isArray(dataset)) {
        dataset = dataset.find((item) => item.id === `--${seriesComponent.measure}`);
    }
    else if (isArray(dataset)) {
        dataset = dataset[0];
    }
    return dataset.dimensions ? indexOf(dataset.dimensions, seriesComponent.measure) : seriesComponent.measure;
}
export function getLegendColorForVisualMap(colors) {
    if (colors && colors.length > 2) {
        return new LinearGradient(0, 0, 1, 0, [
            {
                offset: 0,
                color: colors[0]
            },
            {
                offset: 0.5,
                color: colors[Math.floor(colors.length / 2)]
            },
            {
                offset: 1,
                color: colors[colors.length - 1]
            }
        ]);
    }
    else if (colors && colors.length > 1) {
        return new LinearGradient(0, 0, 1, 0, [
            {
                offset: 0,
                color: colors[0]
            },
            {
                offset: 1,
                color: colors[colors.length - 1]
            }
        ]);
    }
    else if (colors && colors.length > 0) {
        return colors[0];
    }
    return undefined;
}
export function getSymbolSizeVisualMap(measure, property, dataset, visualMaps) {
    // 原来为什么注释掉, 是不是因为 dataset 情况变化
    const _dataset = Array.isArray(dataset) ? dataset[0] : dataset;
    const dataMin = minBy(_dataset.source, measure.measure)?.[measure.measure];
    const dataMax = maxBy(_dataset.source, measure.measure)?.[measure.measure];
    const visualMap = visualMaps?.find((item) => item.type === 'continuous') ?? {};
    return {
        dimension: _dataset.dimensions ? _dataset.dimensions.indexOf(measure.measure) : measure.measure,
        min: dataMin,
        max: dataMax,
        itemWidth: 30,
        itemHeight: 120,
        calculable: true,
        precision: 0.1,
        text: [property.caption || property.name],
        textGap: 30,
        inRange: {
            symbolSize: [10, 70]
        },
        outOfRange: {
            symbolSize: [10, 70],
            color: ['rgba(255,255,255,0.4)']
        },
        controller: {
            inRange: {
                color: ['#c23531']
            },
            outOfRange: {
                color: ['#999']
            }
        },
        ...omitBlank(visualMap)
    };
}
// export function getColorLightnessVisualMap(
//   measure: ChartMeasure,
//   property: Property,
//   dataset: IDataset | IDataset[],
//   visualMaps?: IVisualMap[]
// ) {
//   const _dataset = Array.isArray(dataset) ? dataset[0] : dataset
//   const dataMin = minBy(_dataset.source, measure.measure)?.[measure.measure]
//   const dataMax = maxBy(_dataset.source, measure.measure)?.[measure.measure]
//   const visualMap = visualMaps?.find((item) => item.type === 'continuous') ?? {}
//   return {
//     dimension: _dataset.dimensions ? _dataset.dimensions.indexOf(measure.measure) : measure.measure,
//     min: dataMin,
//     max: dataMax,
//     itemHeight: 120,
//     text: [property.label || property.name],
//     textGap: 30,
//     inRange: {
//       colorLightness: [0.9, 0.5]
//     },
//     outOfRange: {
//       color: ['rgba(255,255,255,0.4)']
//     },
//     controller: {
//       inRange: {
//         color: ['#c23531']
//       },
//       outOfRange: {
//         color: ['#999']
//       }
//     },
//     ...visualMap
//   }
// }
export function measuresToSeriesComponents(measures, data, entityType, settings) {
    const tooltips = measures.filter(({ role }) => role === ChartMeasureRoleType.Tooltip);
    const _measures = measures.filter(({ role }) => !includes([ChartMeasureRoleType.Tooltip, ChartMeasureRoleType.Size, ChartMeasureRoleType.Lightness], role));
    return _measures.map((measure) => {
        const measureName = getPropertyMeasure(measure);
        const measureProperty = getEntityProperty(entityType, measure);
        const valueAxisIndex = measure.role === ChartMeasureRoleType.Axis2 ? 1 : 0;
        const dataNotNull = data.filter((item) => !isNil(item[measureName]));
        const minItem = minBy(dataNotNull, measureName);
        const maxItem = maxBy(dataNotNull, measureName);
        return {
            ...measure,
            id: getPropertyMeasure(measure),
            name: measureProperty?.caption,
            label: measureProperty?.caption,
            seriesType: measure.shapeType,
            property: measureProperty,
            dataMin: minItem?.[measureName],
            dataMax: maxItem?.[measureName],
            dataSize: data.length,
            valueAxisIndex,
            tooltip: tooltips.map(({ measure }) => measure),
            sizeMeasure: measures.find(({ role }) => role === ChartMeasureRoleType.Size),
            lightnessMeasure: measures.find(({ role }) => role === ChartMeasureRoleType.Lightness)
        };
    });
}
export function dimensionToSeriesComponent(name, dimension, measures, data, entityType, settings) {
    const property = getEntityProperty(entityType, dimension);
    const tooltips = measures.filter(({ role }) => role === ChartMeasureRoleType.Tooltip);
    return {
        id: settings?.universalTransition ? getPropertyHierarchy(dimension) : null,
        name: name ?? property.caption ?? property.name,
        label: property?.caption,
        property,
        dataSize: data.length,
        tooltip: tooltips.map(({ measure }) => measure),
        sizeMeasure: measures.find(({ role }) => role === ChartMeasureRoleType.Size),
        lightnessMeasure: measures.find(({ role }) => role === ChartMeasureRoleType.Lightness)
    };
}
/**
 *
 * @param chartAnnotation
 * @param entityType
 * @param items
 * @param chartOptions
 * @param locale
 * @returns
 */
export function getCoordinateSystem(context, items, locale) {
    const { chartAnnotation, entityType, options: chartOptions } = context;
    const measure = chartAnnotation.measures.find((item) => !item.role || item.role === ChartMeasureRoleType.Axis1)?.measure;
    const [categoryOrient, valueOrient] = axisOrient(chartAnnotation.chartType.orient);
    const category = getChartCategory(chartAnnotation);
    const category2 = getChartCategory2(chartAnnotation);
    if (chartAnnotation.chartType.type === 'Scatter') {
        const valueAxises = getValueAxis(chartAnnotation, entityType, chartOptions, locale);
        return {
            categoryAxis: { orient: categoryOrient, axis: [valueAxises[0]] },
            valueAxis: { orient: valueOrient, axis: [valueAxises[1]] }
        };
    }
    else {
        if (!category) {
            throw new Error(`Can'y found category in chart`);
        }
        let valueAxis = null;
        // 设置维度轴值
        const categoryAxis = getCategoryAxis(context, items, category, getEntityHierarchy(entityType, category), chartOptions, measure);
        categoryAxis.orient = categoryOrient;
        if (category2) {
            valueAxis = [
                getCategoryAxis(context, items, category2, getEntityHierarchy(entityType, category2), chartOptions, measure)
            ];
        }
        else {
            valueAxis = getValueAxis(chartAnnotation, entityType, chartOptions, locale);
        }
        valueAxis.orient = valueOrient;
        return {
            categoryAxis: { orient: categoryOrient, axis: [categoryAxis] },
            valueAxis: { orient: valueOrient, axis: valueAxis }
        };
    }
}
export function getPolarCoordinateSystem(context, items, locale) {
    const { chartAnnotation, entityType, options: chartOptions } = context;
    const measure = chartAnnotation.measures.find((item) => !item.role || item.role === ChartMeasureRoleType.Axis1)?.measure;
    const [categoryOrient, valueOrient] = (() => {
        // Chart Orient
        if (chartAnnotation.chartType.orient === ChartOrient.vertical) {
            return [AxisEnum.angle, AxisEnum.radius];
        }
        else {
            return [AxisEnum.radius, AxisEnum.angle];
        }
    })();
    const category = getChartCategory(chartAnnotation);
    const category2 = getChartCategory2(chartAnnotation);
    let valueAxis = null;
    // 设置维度轴值
    const categoryAxis = getCategoryAxis(context, items, category, getEntityHierarchy(entityType, category), chartOptions, measure);
    categoryAxis.orient = categoryOrient;
    if (category2) {
        valueAxis = [
            getCategoryAxis(context, items, category2, getEntityHierarchy(entityType, category2), chartOptions, measure)
        ];
    }
    else {
        valueAxis = getValueAxis(chartAnnotation, entityType, chartOptions, locale);
    }
    valueAxis.orient = valueOrient;
    return {
        categoryAxis: { orient: categoryOrient, axis: [categoryAxis] },
        valueAxis: { orient: valueOrient, axis: valueAxis }
    };
}
//# sourceMappingURL=cartesian.js.map